@@ -1,6 +1,6 @@
 /*
    Copyright (c) 2000, 2011, Oracle and/or its affiliates.
-   Copyright (c) 2020, MariaDB
+   Copyright (c) 2020, 2021, MariaDB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -1124,21 +1124,25 @@ void Global_read_lock::unlock_global_read_lock(THD *thd)
   thd->mdl_context.release_lock(m_mdl_global_read_lock);
 
 #ifdef WITH_WSREP
-  if (m_state == GRL_ACQUIRED_AND_BLOCKS_COMMIT)
+  if (m_state == GRL_ACQUIRED_AND_BLOCKS_COMMIT &&
+      wsrep_locked_seqno != WSREP_SEQNO_UNDEFINED)
   {
     Wsrep_server_state& server_state= Wsrep_server_state::instance();
     if (server_state.state() == Wsrep_server_state::s_donor ||
         (WSREP_NNULL(thd) &&
          server_state.state() != Wsrep_server_state::s_synced))
     {
-      /* TODO: maybe redundant here?: */
-      wsrep_locked_seqno= WSREP_SEQNO_UNDEFINED;
       server_state.resume();
+      wsrep_locked_seqno= WSREP_SEQNO_UNDEFINED;
     }
     else if (WSREP_NNULL(thd) &&
              server_state.state() == Wsrep_server_state::s_synced)
     {
+      THD_STAGE_INFO(thd, stage_waiting_flow);
+      WSREP_DEBUG("unlock_global_read_lock: waiting for flow control for %s",
+                  wsrep_thd_query(thd));
       server_state.resume_and_resync();
+      wsrep_locked_seqno= WSREP_SEQNO_UNDEFINED;
     }
   }
 #endif /* WITH_WSREP */
