@@ -1,5 +1,5 @@
 /* Copyright (c) 2006, 2015, Oracle and/or its affiliates.
-   Copyright (c) 2010, 2022, MariaDB Corporation.
+   Copyright (c) 2010, 2020, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -40,7 +40,7 @@
 #include "ha_partition.h"
 
 
-partition_info *partition_info::get_clone(THD *thd, bool empty_data_and_index_file)
+partition_info *partition_info::get_clone(THD *thd)
 {
   MEM_ROOT *mem_root= thd->mem_root;
   DBUG_ENTER("partition_info::get_clone");
@@ -60,22 +60,21 @@ partition_info *partition_info::get_clone(THD *thd, bool empty_data_and_index_fi
   {
     List_iterator<partition_element> subpart_it(part->subpartitions);
     partition_element *subpart;
-    partition_element *part_clone= new (mem_root) partition_element(*part);
+    partition_element *part_clone= new (mem_root) partition_element();
     if (!part_clone)
       DBUG_RETURN(NULL);
+
+    *part_clone= *part;
     part_clone->subpartitions.empty();
     while ((subpart= (subpart_it++)))
     {
-      partition_element *subpart_clone= new (mem_root) partition_element(*subpart);
+      partition_element *subpart_clone= new (mem_root) partition_element();
       if (!subpart_clone)
         DBUG_RETURN(NULL);
-      if (empty_data_and_index_file)
-        subpart_clone->data_file_name= subpart_clone->index_file_name= NULL;
+
+      *subpart_clone= *subpart;
       part_clone->subpartitions.push_back(subpart_clone, mem_root);
     }
-
-    if (empty_data_and_index_file)
-      part_clone->data_file_name= part_clone->index_file_name= NULL;
     clone->partitions.push_back(part_clone, mem_root);
     part_clone->list_val_list.empty();
     List_iterator<part_elem_value> list_val_it(part->list_val_list);
@@ -127,7 +126,7 @@ partition_info *partition_info::get_clone(THD *thd, bool empty_data_and_index_fi
 /**
   Mark named [sub]partition to be used/locked.
 
-  @param part_name  Partition name to match.
+  @param part_name  Partition name to match.  Must be \0 terminated!
   @param length     Partition name length.
 
   @return Success if partition found
@@ -141,7 +140,10 @@ bool partition_info::add_named_partition(const char *part_name, size_t length)
   PART_NAME_DEF *part_def;
   Partition_share *part_share;
   DBUG_ENTER("partition_info::add_named_partition");
-  DBUG_ASSERT(table && table->s && table->s->ha_share);
+  DBUG_ASSERT(part_name[length] == 0);
+  DBUG_ASSERT(table);
+  DBUG_ASSERT(table->s);
+  DBUG_ASSERT(table->s->ha_share);
   part_share= static_cast<Partition_share*>((table->s->ha_share));
   DBUG_ASSERT(part_share->partition_name_hash_initialized);
   part_name_hash= &part_share->partition_name_hash;
@@ -173,9 +175,9 @@ bool partition_info::add_named_partition(const char *part_name, size_t length)
     else
       bitmap_set_bit(&read_partitions, part_def->part_id);
   }
-  DBUG_PRINT("info", ("Found partition %u is_subpart %d for name %s",
+  DBUG_PRINT("info", ("Found partition %u is_subpart %d for name %.*s",
                       part_def->part_id, part_def->is_subpart,
-                      part_name));
+                      length, part_name));
   DBUG_RETURN(false);
 }
 
@@ -817,41 +819,42 @@ bool partition_info::has_unique_name(partition_element *element)
     vers_info->interval   Limit by fixed time interval
     vers_info->hist_part  (out) Working history partition
 */
-int partition_info::vers_set_hist_part(THD *thd)
+void partition_info::vers_set_hist_part(THD *thd)
 {
-  if (!vers_require_hist_part(thd))
-    return 0;
-
-  if (table->pos_in_table_list &&
-      table->pos_in_table_list->partition_names)
-  {
-    return HA_ERR_PARTITION_LIST;
-  }
   if (vers_info->limit)
   {
     ha_partition *hp= (ha_partition*)(table->file);
-    partition_element *next;
+    partition_element *next= NULL;
     List_iterator<partition_element> it(partitions);
-    ha_rows records= 0;
-    vers_info->hist_part= partitions.head();
+    while (next != vers_info->hist_part)
+      next= it++;
+    ha_rows records= hp->part_records(next);
     while ((next= it++) != vers_info->now_part)
     {
-      DBUG_ASSERT(bitmap_is_set(&read_partitions, next->id));
       ha_rows next_records= hp->part_records(next);
       if (next_records == 0)
         break;
       vers_info->hist_part= next;
       records= next_records;
     }
-    if (records >= vers_info->limit && next != vers_info->now_part)
-      vers_info->hist_part= next;
-    return 0;
+    if (records >= vers_info->limit)
+    {
+      if (next == vers_info->now_part)
+      {
+        my_error(WARN_VERS_PART_FULL, MYF(ME_WARNING|ME_ERROR_LOG),
+                table->s->db.str, table->s->table_name.str,
+                vers_info->hist_part->partition_name, "LIMIT");
+      }
+      else
+        vers_info->hist_part= next;
+    }
+    return;
   }
 
   if (vers_info->interval.is_set())
   {
     if (vers_info->hist_part->range_value > thd->query_start())
-      return 0;
+      return;
 
     partition_element *next= NULL;
     List_iterator<partition_element> it(partitions);
@@ -862,49 +865,9 @@ int partition_info::vers_set_hist_part(THD *thd)
     {
       vers_info->hist_part= next;
       if (next->range_value > thd->query_start())
-        return 0;
+        return;
     }
   }
-  return 0;
-}
-
-
-/**
-  Warn at the end of DML command if the last history partition is out of LIMIT.
-*/
-void partition_info::vers_check_limit(THD *thd)
-{
-  if (!vers_info->limit ||
-      vers_info->hist_part->id + 1 < vers_info->now_part->id)
-    return;
-
-  /*
-    NOTE: at this point read_partitions bitmap is already pruned by DML code,
-    we have to set read bits for working history partition. We could use
-    bitmap_set_all(), but this is not optimal since there can be quite a number
-    of partitions.
-  */
-#ifndef DBUG_OFF
-  const uint32 sub_factor= num_subparts ? num_subparts : 1;
-  uint32 part_id= vers_info->hist_part->id * sub_factor;
-  const uint32 part_id_end= part_id + sub_factor;
-  DBUG_ASSERT(part_id_end <= num_parts * sub_factor);
-#endif
-
-  ha_partition *hp= (ha_partition*)(table->file);
-  ha_rows hist_rows= hp->part_records(vers_info->hist_part);
-  if (hist_rows >= vers_info->limit)
-  {
-    push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
-                        WARN_VERS_PART_FULL,
-                        ER_THD(thd, WARN_VERS_PART_FULL),
-                        table->s->db.str, table->s->table_name.str,
-                        vers_info->hist_part->partition_name, "LIMIT");
-
-    sql_print_warning(ER_THD(thd, WARN_VERS_PART_FULL),
-                      table->s->db.str, table->s->table_name.str,
-                      vers_info->hist_part->partition_name, "LIMIT");
-  }
 }
 
 
@@ -1523,6 +1486,7 @@ bool partition_info::set_up_charset_field_preps(THD *thd)
   uchar **char_ptrs;
   unsigned i;
   size_t size;
+  uint tot_fields= 0;
   uint tot_part_fields= 0;
   uint tot_subpart_fields= 0;
   DBUG_ENTER("set_up_charset_field_preps");
@@ -1534,8 +1498,13 @@ bool partition_info::set_up_charset_field_preps(THD *thd)
     ptr= part_field_array;
     /* Set up arrays and buffers for those fields */
     while ((field= *(ptr++)))
+    {
       if (field_is_partition_charset(field))
+      {
         tot_part_fields++;
+        tot_fields++;
+      }
+    }
     size= tot_part_fields * sizeof(char*);
     if (!(char_ptrs= (uchar**)thd->calloc(size)))
       goto error;
@@ -1569,8 +1538,13 @@ bool partition_info::set_up_charset_field_preps(THD *thd)
     /* Set up arrays and buffers for those fields */
     ptr= subpart_field_array;
     while ((field= *(ptr++)))
+    {
       if (field_is_partition_charset(field))
+      {
         tot_subpart_fields++;
+        tot_fields++;
+      }
+    }
     size= tot_subpart_fields * sizeof(char*);
     if (!(char_ptrs= (uchar**) thd->calloc(size)))
       goto error;
@@ -2452,7 +2426,7 @@ bool partition_info::has_same_partitioning(partition_info *new_part_info)
       partition_element *new_part_elem= new_part_it++;
       /*
         The following must match:
-        partition_name, tablespace_name, data_file_name, index_file_name,
+        partition_name, data_file_name, index_file_name,
         engine_type, part_max_rows, part_min_rows, nodegroup_id.
         (max_value, signed_flag, has_null_value only on partition level,
         RANGE/LIST)
@@ -2538,9 +2512,7 @@ bool partition_info::has_same_partitioning(partition_info *new_part_info)
             if (strcmp_null(sub_part_elem->data_file_name,
                             new_sub_part_elem->data_file_name) ||
                 strcmp_null(sub_part_elem->index_file_name,
-                            new_sub_part_elem->index_file_name) ||
-                strcmp_null(sub_part_elem->tablespace_name,
-                            new_sub_part_elem->tablespace_name))
+                            new_sub_part_elem->index_file_name))
               DBUG_RETURN(false);
 
           } while (++j < num_subparts);
@@ -2556,9 +2528,7 @@ bool partition_info::has_same_partitioning(partition_info *new_part_info)
         if (strcmp_null(part_elem->data_file_name,
                         new_part_elem->data_file_name) ||
             strcmp_null(part_elem->index_file_name,
-                        new_part_elem->index_file_name) ||
-            strcmp_null(part_elem->tablespace_name,
-                        new_part_elem->tablespace_name))
+                        new_part_elem->index_file_name))
           DBUG_RETURN(false);
       }
     } while (++i < num_parts);
