@@ -33,6 +33,7 @@
 #include "sql_statistics.h"
 #include "wsrep_mysqld.h"
 
+const LEX_CSTRING msg_status= {STRING_WITH_LEN("status")};
 const LEX_CSTRING msg_repair= { STRING_WITH_LEN("repair") };
 const LEX_CSTRING msg_assign_to_keycache=
 { STRING_WITH_LEN("assign_to_keycache") };
@@ -476,19 +477,66 @@ static bool wsrep_toi_replication(THD *thd, TABLE_LIST *tables)
   /* now TOI replication, with no locks held */
   if (keys.empty())
   {
-    WSREP_TO_ISOLATION_BEGIN_WRTCHK(NULL, NULL, tables);
-  } else {
-    WSREP_TO_ISOLATION_BEGIN_FK_TABLES(NULL, NULL, tables, &keys) {
+    if (!thd->lex->no_write_to_binlog &&
+	wsrep_to_isolation_begin(thd, NULL, NULL, tables))
+      return true;
+  }
+  else
+  {
+    if (!thd->lex->no_write_to_binlog &&
+	wsrep_to_isolation_begin(thd, NULL, NULL, tables, NULL, &keys))
       return true;
-    }
   }
   return false;
-
- wsrep_error_label:
-  return true;
 }
 #endif /* WITH_WSREP */
 
+
+static void send_read_only_warning(THD *thd, const LEX_CSTRING *msg_status,
+                                   const LEX_CSTRING *table_name)
+{
+  Protocol *protocol= thd->protocol;
+  char buf[MYSQL_ERRMSG_SIZE];
+  size_t length;
+  length= my_snprintf(buf, sizeof(buf),
+                      ER_THD(thd, ER_OPEN_AS_READONLY),
+                      table_name->str);
+  protocol->store(msg_status, system_charset_info);
+  protocol->store(buf, length, system_charset_info);
+}
+
+
+/**
+  Collect field names of result set that will be sent to a client
+
+  @param      thd     Thread data object
+  @param[out] fields  List of fields whose metadata should be collected for
+                      sending to client
+*/
+
+void fill_check_table_metadata_fields(THD *thd, List<Item>* fields)
+{
+  Item *item;
+
+  item= new (thd->mem_root) Item_empty_string(thd, "Table", NAME_CHAR_LEN * 2);
+  item->set_maybe_null();
+  fields->push_back(item, thd->mem_root);
+
+  item= new (thd->mem_root) Item_empty_string(thd, "Op", 10);
+  item->set_maybe_null();
+  fields->push_back(item, thd->mem_root);
+
+  item= new (thd->mem_root) Item_empty_string(thd, "Msg_type", 10);
+  item->set_maybe_null();
+  fields->push_back(item, thd->mem_root);
+
+  item= new (thd->mem_root) Item_empty_string(thd, "Msg_text",
+                                              SQL_ADMIN_MSG_TEXT_SIZE);
+  item->set_maybe_null();
+  fields->push_back(item, thd->mem_root);
+}
+
+
 /*
   RETURN VALUES
     FALSE Message sent to net (admin operation went ok)
@@ -512,7 +560,6 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
 {
   TABLE_LIST *table;
   List<Item> field_list;
-  Item *item;
   Protocol *protocol= thd->protocol;
   LEX *lex= thd->lex;
   int result_code;
@@ -524,21 +571,8 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
   DBUG_ENTER("mysql_admin_table");
   DBUG_PRINT("enter", ("extra_open_options: %u", extra_open_options));
 
-  field_list.push_back(item= new (thd->mem_root)
-                       Item_empty_string(thd, "Table",
-                                         NAME_CHAR_LEN * 2), thd->mem_root);
-  item->set_maybe_null();
-  field_list.push_back(item= new (thd->mem_root)
-                       Item_empty_string(thd, "Op", 10), thd->mem_root);
-  item->set_maybe_null();
-  field_list.push_back(item= new (thd->mem_root)
-                       Item_empty_string(thd, "Msg_type", 10), thd->mem_root);
-  item->set_maybe_null();
-  field_list.push_back(item= new (thd->mem_root)
-                       Item_empty_string(thd, "Msg_text",
-                                         SQL_ADMIN_MSG_TEXT_SIZE),
-                       thd->mem_root);
-  item->set_maybe_null();
+  fill_check_table_metadata_fields(thd, &field_list);
+
   if (protocol->send_result_set_metadata(&field_list,
                             Protocol::SEND_NUM_ROWS | Protocol::SEND_EOF))
     DBUG_RETURN(TRUE);
@@ -605,6 +639,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
 
     if (thd->check_killed())
     {
+      open_error= false;
       fatal_error= true;
       result_code= HA_ADMIN_FAILED;
       goto send_result;
@@ -678,8 +713,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
             protocol->store(operator_name, system_charset_info);
             protocol->store(&error_clex_str, system_charset_info);
             length= my_snprintf(buff, sizeof(buff),
-                                ER_THD(thd, ER_DROP_PARTITION_NON_EXISTENT),
-                                table_name.str);
+                                ER_THD(thd, ER_PARTITION_DOES_NOT_EXIST));
             protocol->store(buff, length, system_charset_info);
             if(protocol->write())
               goto err;
@@ -755,20 +789,16 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
       goto send_result;
     }
 
-    if ((table->table->db_stat & HA_READ_ONLY) && open_for_modify)
+    if ((table->table->db_stat & HA_READ_ONLY) && open_for_modify &&
+        operator_func != &handler::ha_analyze)
     {
       /* purecov: begin inspected */
-      char buff[FN_REFLEN + MYSQL_ERRMSG_SIZE];
-      size_t length;
       enum_sql_command save_sql_command= lex->sql_command;
       DBUG_PRINT("admin", ("sending error message"));
       protocol->prepare_for_resend();
       protocol->store(&table_name, system_charset_info);
       protocol->store(operator_name, system_charset_info);
-      protocol->store(&error_clex_str, system_charset_info);
-      length= my_snprintf(buff, sizeof(buff), ER_THD(thd, ER_OPEN_AS_READONLY),
-                          table_name.str);
-      protocol->store(buff, length, system_charset_info);
+      send_read_only_warning(thd, &error_clex_str, &table_name);
       trans_commit_stmt(thd);
       trans_commit(thd);
       close_thread_tables(thd);
@@ -904,6 +934,15 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
 
     if (compl_result_code == HA_ADMIN_OK && collect_eis)
     {
+      if (result_code == HA_ERR_TABLE_READONLY)
+      {
+        protocol->prepare_for_resend();
+        protocol->store(&table_name, system_charset_info);
+        protocol->store(operator_name, system_charset_info);
+        send_read_only_warning(thd, &msg_status, &table_name);
+        (void) protocol->write();
+        result_code= HA_ADMIN_OK;
+      }
       /*
         Here we close and reopen table in read mode because operation of
         collecting statistics is long and it will be better do not block
@@ -939,7 +978,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
             enum enum_field_types type= (*field_ptr)->type();
             if (type < MYSQL_TYPE_MEDIUM_BLOB ||
                 type > MYSQL_TYPE_BLOB)
-              bitmap_set_bit(tab->read_set, fields);
+              tab->field[fields]->register_field_in_read_map();
             else
               push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
                                   ER_NO_EIS_FOR_FIELD,
@@ -967,7 +1006,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
             enum enum_field_types type= tab->field[pos]->type();
             if (type < MYSQL_TYPE_MEDIUM_BLOB ||
                 type > MYSQL_TYPE_BLOB)
-              bitmap_set_bit(tab->read_set, pos);
+              tab->field[pos]->register_field_in_read_map();
             else
               push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
                                   ER_NO_EIS_FOR_FIELD,
@@ -1013,7 +1052,7 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
         protocol->prepare_for_resend();
         protocol->store(&table_name, system_charset_info);
         protocol->store(operator_name, system_charset_info);
-        protocol->store(STRING_WITH_LEN("status"), system_charset_info);
+        protocol->store(&msg_status, system_charset_info);
 	protocol->store(STRING_WITH_LEN("Engine-independent statistics collected"), 
                         system_charset_info);
         if (protocol->write())
@@ -1083,25 +1122,25 @@ send_result_message:
       break;
 
     case HA_ADMIN_OK:
-      protocol->store(STRING_WITH_LEN("status"), system_charset_info);
+      protocol->store(&msg_status, system_charset_info);
       protocol->store(STRING_WITH_LEN("OK"), system_charset_info);
       break;
 
     case HA_ADMIN_FAILED:
-      protocol->store(STRING_WITH_LEN("status"), system_charset_info);
+      protocol->store(&msg_status, system_charset_info);
       protocol->store(STRING_WITH_LEN("Operation failed"),
                       system_charset_info);
       break;
 
     case HA_ADMIN_REJECT:
-      protocol->store(STRING_WITH_LEN("status"), system_charset_info);
+      protocol->store(&msg_status, system_charset_info);
       protocol->store(STRING_WITH_LEN("Operation need committed state"),
                       system_charset_info);
       open_for_modify= FALSE;
       break;
 
     case HA_ADMIN_ALREADY_DONE:
-      protocol->store(STRING_WITH_LEN("status"), system_charset_info);
+      protocol->store(&msg_status, system_charset_info);
       protocol->store(STRING_WITH_LEN("Table is already up to date"),
                       system_charset_info);
       break;
@@ -1246,7 +1285,11 @@ send_result_message:
       fatal_error=1;
       break;
     }
-
+    case HA_ERR_TABLE_READONLY:
+    {
+      send_read_only_warning(thd, &msg_status, &table_name);
+      break;
+    }
     default:				// Probably HA_ADMIN_INTERNAL_ERROR
       {
         char buf[MYSQL_ERRMSG_SIZE];
