@@ -259,11 +259,11 @@ Alter_table_ctx::Alter_table_ctx()
     tables_opened(0),
     db(null_clex_str), table_name(null_clex_str), alias(null_clex_str),
     new_db(null_clex_str), new_name(null_clex_str), new_alias(null_clex_str),
+    storage_engine_partitioned(false),
+    tmp_storage_engine_name_partitioned(false),
     fk_error_if_delete_row(false), fk_error_id(NULL),
-    fk_error_table(NULL)
-#ifdef DBUG_ASSERT_EXISTS
-    , tmp_table(false)
-#endif
+    fk_error_table(NULL),
+    tmp_table(false)
 {
 }
 
@@ -279,11 +279,10 @@ Alter_table_ctx::Alter_table_ctx(THD *thd, TABLE_LIST *table_list,
   : implicit_default_value_error_field(NULL), error_if_not_empty(false),
     tables_opened(tables_opened_arg),
     new_db(*new_db_arg), new_name(*new_name_arg),
+    tmp_storage_engine_name_partitioned(false),
     fk_error_if_delete_row(false), fk_error_id(NULL),
-    fk_error_table(NULL)
-#ifdef DBUG_ASSERT_EXISTS
-    , tmp_table(false)
-#endif
+    fk_error_table(NULL),
+    tmp_table(false)
 {
   /*
     Assign members db, table_name, new_db and new_name
@@ -362,10 +361,22 @@ Alter_table_ctx::Alter_table_ctx(THD *thd, TABLE_LIST *table_list,
       this case. This fact is enforced with assert.
     */
     build_tmptable_filename(thd, tmp_path, sizeof(tmp_path));
-#ifdef DBUG_ASSERT_EXISTS
     tmp_table= true;
-#endif
   }
+  if ((id.length= table_list->table->s->tabledef_version.length))
+    memcpy(id_buff, table_list->table->s->tabledef_version.str, MY_UUID_SIZE);
+  id.str= id_buff;
+  storage_engine_partitioned= table_list->table->file->partition_engine();
+  storage_engine_name.str= storage_engine_buff;
+  storage_engine_name.length= ((strmake(storage_engine_buff,
+                                        table_list->table->file->
+                                        real_table_type(),
+                                        sizeof(storage_engine_buff)-1)) -
+                               storage_engine_buff);
+  tmp_storage_engine_name.str= tmp_storage_engine_buff;
+  tmp_storage_engine_name.length= 0;
+  tmp_id.str= 0;
+  tmp_id.length= 0;
 }
 
 
@@ -429,6 +440,8 @@ bool Sql_cmd_alter_table::execute(THD *thd)
     as for RENAME TO, as being done by SQLCOM_RENAME_TABLE
   */
   if ((alter_info.partition_flags & ALTER_PARTITION_DROP) ||
+      (alter_info.partition_flags & ALTER_PARTITION_CONVERT_IN) ||
+      (alter_info.partition_flags & ALTER_PARTITION_CONVERT_OUT) ||
       (alter_info.flags & ALTER_RENAME))
     priv_needed|= DROP_ACL;
 
