@@ -256,7 +256,7 @@ buf_dump(
 #ifdef _WIN32
 	/* use my_fopen() for correct permissions during bootstrap*/
 	f = my_fopen(tmp_filename, O_RDWR|O_TRUNC|O_CREAT, 0);
-#elif defined(__GLIBC__) || defined(__WIN__) || O_CLOEXEC == 0
+#elif defined(__GLIBC__) || O_CLOEXEC == 0
 	f = fopen(tmp_filename, "w" STR_O_CLOEXEC);
 #else
 	{
@@ -634,7 +634,7 @@ buf_load()
 	/* Avoid calling the expensive fil_space_t::get() for each
 	page within the same tablespace. dump[] is sorted by (space, page),
 	so all pages from a given tablespace are consecutive. */
-	ulint		cur_space_id = dump[0].space();
+	uint32_t	cur_space_id = dump[0].space();
 	fil_space_t*	space = fil_space_t::get(cur_space_id);
 	ulint		zip_size = space ? space->zip_size() : 0;
 
@@ -646,10 +646,9 @@ buf_load()
 	for (i = 0; i < dump_n && !SHUTTING_DOWN(); i++) {
 
 		/* space_id for this iteration of the loop */
-		const ulint	this_space_id = dump[i].space();
+		const uint32_t this_space_id = dump[i].space();
 
-		if (this_space_id == SRV_TMP_SPACE_ID) {
-			/* Ignore the innodb_temporary tablespace. */
+		if (this_space_id >= SRV_SPACE_ID_UPPER_BOUND) {
 			continue;
 		}
 
@@ -684,7 +683,7 @@ buf_load()
 		}
 
 		space->reacquire();
-		buf_read_page_background(space, dump[i], zip_size, true);
+		buf_read_page_background(space, dump[i], zip_size);
 
 		if (buf_load_abort_flag) {
 			if (space) {
@@ -721,6 +720,10 @@ buf_load()
 
 	ut_free(dump);
 
+	if (i == dump_n) {
+		os_aio_wait_until_no_pending_reads();
+	}
+
 	ut_sprintf_timestamp(now);
 
 	if (i == dump_n) {
