@@ -267,20 +267,20 @@ Datafile::read_first_page(bool read_only_mode)
 			IORequestReadPartial, m_handle, m_first_page, 0,
 			page_size, &n_read);
 
-		if (err == DB_IO_ERROR && n_read >= UNIV_PAGE_SIZE_MIN) {
-
-			page_size >>= 1;
-
-		} else if (err == DB_SUCCESS) {
-
+		if (err == DB_SUCCESS) {
 			ut_a(n_read == page_size);
-
 			break;
+		}
 
+		if (err == DB_IO_ERROR && n_read == 0) {
+			break;
+		}
+		if (err == DB_IO_ERROR && n_read >= UNIV_PAGE_SIZE_MIN) {
+			page_size >>= 1;
 		} else if (srv_operation == SRV_OPERATION_BACKUP) {
 			break;
 		} else {
-			ib::error() << "Cannot read first page of '"
+			ib::info() << "Cannot read first page of '"
 				<< m_filepath << "': " << err;
 			break;
 		}
@@ -304,8 +304,8 @@ Datafile::read_first_page(bool read_only_mode)
 					      + m_first_page);
 		m_flags = fsp_header_get_flags(m_first_page);
 		if (!fil_space_t::is_valid_flags(m_flags, m_space_id)) {
-			ulint cflags = fsp_flags_convert_from_101(m_flags);
-			if (cflags == ULINT_UNDEFINED) {
+			uint32_t cflags = fsp_flags_convert_from_101(m_flags);
+			if (cflags == UINT32_MAX) {
 				ib::error()
 					<< "Invalid flags " << ib::hex(m_flags)
 					<< " in " << m_filepath;
@@ -342,8 +342,7 @@ in order for this function to validate it.
 @param[in]	flags		The expected tablespace flags.
 @retval DB_SUCCESS if tablespace is valid, DB_ERROR if not.
 m_is_valid is also set true on success, else false. */
-dberr_t
-Datafile::validate_to_dd(ulint space_id, ulint flags)
+dberr_t Datafile::validate_to_dd(uint32_t space_id, uint32_t flags)
 {
 	dberr_t err;
 
@@ -400,10 +399,19 @@ Datafile::validate_for_recovery()
 	err = validate_first_page(0);
 
 	switch (err) {
-	case DB_SUCCESS:
 	case DB_TABLESPACE_EXISTS:
 		break;
-
+	case DB_SUCCESS:
+		if (!m_defer || !m_space_id) {
+			break;
+		}
+		/* InnoDB should check whether the deferred
+		tablespace page0 can be recovered from
+		double write buffer. InnoDB should try
+	        to recover only if m_space_id exists because
+		dblwr pages can be searched via {space_id, 0}.
+		m_space_id is set in read_first_page(). */
+		/* fall through */
 	default:
 		/* Re-open the file in read-write mode  Attempt to restore
 		page 0 from doublewrite and read the space ID from a survey
@@ -414,23 +422,30 @@ Datafile::validate_for_recovery()
 			return(err);
 		}
 
-		err = find_space_id();
-		if (err != DB_SUCCESS || m_space_id == 0) {
-			ib::error() << "Datafile '" << m_filepath << "' is"
-				" corrupted. Cannot determine the space ID from"
-				" the first 64 pages.";
-			return(err);
+		if (!m_defer) {
+			err = find_space_id();
+			if (err != DB_SUCCESS || m_space_id == 0) {
+				ib::error() << "Datafile '" << m_filepath
+					<< "' is corrupted. Cannot determine "
+					"the space ID from the first 64 pages.";
+				return(err);
+			}
 		}
-		if (m_space_id == ULINT_UNDEFINED) {
+
+		if (m_space_id == UINT32_MAX) {
 			return DB_SUCCESS; /* empty file */
 		}
 
 		if (restore_from_doublewrite()) {
+			if (m_defer) {
+				return err;
+			}
 			return(DB_CORRUPTION);
 		}
 
 		/* Free the previously read first page and then re-validate. */
 		free_first_page();
+		m_defer = false;
 		err = validate_first_page(0);
 	}
 
@@ -613,15 +628,15 @@ Datafile::find_space_id()
 	     page_size <<= 1) {
 		/* map[space_id] = count of pages */
 		typedef std::map<
-			ulint,
-			ulint,
-			std::less<ulint>,
-			ut_allocator<std::pair<const ulint, ulint> > >
+			uint32_t,
+			uint32_t,
+			std::less<uint32_t>,
+			ut_allocator<std::pair<const uint32_t, uint32_t> > >
 			Pages;
 
 		Pages	verify;
-		ulint	page_count = 64;
-		ulint	valid_pages = 0;
+		uint32_t page_count = 64;
+		uint32_t valid_pages = 0;
 
 		/* Adjust the number of pages to analyze based on file size */
 		while ((page_count * page_size) > file_size) {
@@ -635,14 +650,14 @@ Datafile::find_space_id()
 		byte*	page = static_cast<byte*>(
 			aligned_malloc(page_size, page_size));
 
-		ulint fsp_flags;
+		uint32_t fsp_flags;
 		/* provide dummy value if the first os_file_read() fails */
 		switch (srv_checksum_algorithm) {
 		case SRV_CHECKSUM_ALGORITHM_STRICT_FULL_CRC32:
 		case SRV_CHECKSUM_ALGORITHM_FULL_CRC32:
 			fsp_flags = 1U << FSP_FLAGS_FCRC32_POS_MARKER
 				| FSP_FLAGS_FCRC32_PAGE_SSIZE()
-				| innodb_compression_algorithm
+				| uint(innodb_compression_algorithm)
 				       << FSP_FLAGS_FCRC32_POS_COMPRESSED_ALGO;
 			break;
 		default:
@@ -682,7 +697,7 @@ Datafile::find_space_id()
 
 			if (noncompressed_ok || compressed_ok) {
 
-				ulint	space_id = mach_read_from_4(page
+				uint32_t space_id = mach_read_from_4(page
 					+ FIL_PAGE_SPACE_ID);
 
 				if (space_id > 0) {
@@ -761,14 +776,14 @@ Datafile::restore_from_doublewrite()
 		return(true);
 	}
 
-	ulint	flags = mach_read_from_4(
+	uint32_t flags = mach_read_from_4(
 		FSP_HEADER_OFFSET + FSP_SPACE_FLAGS + page);
 
 	if (!fil_space_t::is_valid_flags(flags, m_space_id)) {
 		flags = fsp_flags_convert_from_101(flags);
 		/* recv_dblwr_t::validate_page() inside find_page()
 		checked this already. */
-		ut_ad(flags != ULINT_UNDEFINED);
+		ut_ad(flags != UINT32_MAX);
 		/* The flags on the page should be converted later. */
 	}
 
@@ -805,10 +820,12 @@ static char *read_link_file(const char *link_filepath)
     {
       /* Trim whitespace from end of filepath */
       len--;
-      while (filepath[len] >= 0 && filepath[len] <= 0x20)
+      while (static_cast<byte>(filepath[len]) <= 0x20)
+      {
+        if (!len)
+          return nullptr;
         filepath[len--]= 0;
-      if (!*filepath)
-        return nullptr;
+      }
       /* Ensure that the last 2 path separators are forward slashes,
       because elsewhere we are assuming that tablespace file names end
       in "/databasename/tablename.ibd". */
@@ -837,7 +854,7 @@ open that file, and read the contents into m_filepath.
 @param name   table name
 @return filepath()
 @retval nullptr  if the .isl file does not exist or cannot be read */
-const char *RemoteDatafile::open_link_file(const table_name_t &name)
+const char *RemoteDatafile::open_link_file(const fil_space_t::name_type name)
 {
   if (!m_link_filepath)
     m_link_filepath= fil_make_filepath(nullptr, name, ISL, false);
